#uBlock0-legacy-button {
    list-style-image: url('../img/browsericons/icon24.svg');
}
#uBlock0-legacy-button.off {
    list-style-image: url('../img/browsericons/icon24-off.svg');
}

toolbar[iconsize="small"] #uBlock0-legacy-button {
    list-style-image: url('../img/browsericons/icon16.svg');
}
toolbar[iconsize="small"] #uBlock0-legacy-button.off {
    list-style-image: url('../img/browsericons/icon16-off.svg');
}
#uBlock0-legacy-button[badge]::before {
    background: #555;
    color: #fff;
    content: attr(badge);
    font: bold 10px sans-serif;
    margin-top: -2px;
    padding: 0 2px;
    position: fixed;
}
/* This hack required because if the before content changes it de-pops the
   popup (without firing any events). So just hide it instead. Note, can't
   actually *hide* it, or the same thing happens.
**/
#uBlock0-legacy-button[badge=""]::before {
    padding: 0;
}

/* Override off state when in palette */
toolbarpaletteitem #uBlock0-legacy-button.off {
    list-style-image: url('../img/browsericons/icon24.svg');
}

/* Override badge when in palette */
toolbarpaletteitem #uBlock0-legacy-button[badge]::before {
    content: none;
}

/* Prevent pale moon from showing the arrow underneath the button */
/* https://github.com/chrisaljoudi/uBlock/issues/1449#issuecomment-112112761 */
#uBlock0-legacy-button .toolbarbutton-menu-dropmarker {
    display: none;
    -moz-box-orient: horizontal;
}
