ul {
    padding-left: 1em;
    }
ul#userSettings {
    padding-left: 0;
    }
li {
    margin-top: 0.25em;
    }
ul#userSettings,
ul#userSettings ul {
    list-style-type: none;
    }
ul#userSettings .subgroup {
    margin-top: 1em;
    }
ul#userSettings .subgroup > span {
    font-size: larger;
    font-weight: bold;
    }
#advanced-user-enabled ~ a.fa {
    display: none;
    }
body.advancedUser #advanced-user-enabled ~ a.fa {
    display: inline;
    }
#localData > ul > li {
    margin-top: 1em;
    }
#localData > ul > li > ul > li:nth-of-type(2) {
    font-family: monospace;
    }
