html, body {
    margin: 0;
    border: 0;
    padding: 0;
    position: relative;
    width: 100%;
    height: 100%;
    overflow: hidden;
    }
@media screen and (max-device-width: 960px) {
    #dashboard-nav {
        zoom: 1.2;
        }
    }
#dashboard-nav {
    border: 0;
    margin: 0;
    padding: 0;
    top: 0;
    width: 100%;
    z-index: 10;
    }
#dashboard-nav-widgets {
    margin: 0;
    border-bottom: 1px solid #ccc;
    padding: 4px 0 0;
    white-space: nowrap;
    background-color: white;
    }
#dashboard-nav-widgets span {
    padding: 0 0.5em;
    font-size: larger;
    }
.tabButton {
    background-color: #eee;
    border: 1px solid #ccc;
    border-top-left-radius: 3px;
    border-top-right-radius: 3px;
    border-bottom: 1px solid #ccc;
    color: black;
    cursor: pointer;
    display: inline-block;
    font-size: 110%;
    margin: 0 0.2em 0 0;
    padding: 4px;
    position: relative;
    text-decoration: none;
    top: 1px;
    }
.tabButton:focus {
    outline: 0;
    }
.tabButton:active,.tabButton:visited {
    color: inherit;
    }
.tabButton.selected {
    background-color: white;
    border-bottom: 1px solid white;
    }
iframe {
    margin: 0;
    border: 0;
    padding: 0;
    background-color: transparent;
    width: 100%;
    }

@media (max-width: 600px) {
    #dashboard-nav {
        position: relative;
        }
    #dashboard-nav-widgets {
        padding: 1px 0 0 0;
        white-space: normal;
        }
    [data-i18n="extName"] {
        display: none;
        }
    .tabButton {
        border-radius: 0;
        display: inline-block;
        font-size: 100%;
        margin-bottom: 1px;
        margin-right: 1px;
        top: 0;
        }
    .tabButton.selected {
        border-bottom: 1px solid #ccc;
        }
    }
