@font-face {
    font-family: 'FontAwesome';
    font-weight: normal;
    font-style: normal;
    src: url('fonts/fontawesome-webfont.ttf') format('truetype');
    }
.fa {
    display: inline-block;
    font-family: FontAwesome;
    font-style: normal;
    font-weight: normal;
    line-height: 1;
    user-select: none;
      -moz-user-select: none;
      -webkit-user-select: none;
    vertical-align: middle;
    }
body {
    background-color: white;
    color: black;
    font: 14px sans-serif;
    }
textarea {
    font-size: 90%;
    }
/* I designed the button with: http://charliepark.org/bootstrap_buttons/ */
button.custom {
    padding: 0.6em 1em;
    border: 1px solid transparent;
    border-color: #ccc #ccc #bbb #bbb;
    border-radius: 3px;
    background-color: hsl(216, 0%, 75%);
    background-image: linear-gradient(#f2f2f2, #dddddd);
    background-repeat: repeat-x;
    color: #000;
    opacity: 0.8;
    }
button.custom.disabled,
button.custom[disabled] {
    border-color: #ddd #ddd hsl(36, 0%, 85%);
    background-color: hsl(36, 0%, 72%);
    background-image: linear-gradient(#f2f2f2, #dddddd);
    color: #666;
    opacity: 0.6;
    pointer-events: none;
    }
button.custom:hover {
    opacity: 1.0;
    }
button.important {
    padding: 0.6em 1em;
    border: 1px solid transparent;
    border-color: #ffcc7f #ffcc7f hsl(36, 100%, 73%);
    border-radius: 3px;
    background-color: hsl(36, 100%, 75%);
    background-image: linear-gradient(#ffdca8, #ffcc7f);
    background-repeat: repeat-x;
    color: #222;
    opacity: 0.8;
    }
button.important:hover {
    opacity: 1.0;
    }
.hiddenFileInput {
    height: 0;
    visibility: hidden;
    width: 0;
    }
