div > p:first-child {
    margin-top: 0;
    }
div > p:last-child {
    margin-bottom: 0;
    }
code {
    background-color: #eee;
    font: 11px monospace;
    padding: 2px 4px;
}
#diff {
    border: 0;
    border-top: 1px solid #eee;
    margin: 0;
    padding: 0.5em 0 0 0;
    white-space: nowrap;
}
#diff .pane {
    border: 0;
    box-sizing: border-box;
    display: inline-block;
    font: 90%/180% "Noto Mono",monospace;
    margin: 0;
    padding: 0;
    position: relative;
    white-space: normal;
    width: 50%;
    }
#diff .pane .rulesContainer {
    position: relative;
    min-height: 150px; /* too short is confusing */
    }
#diff .ruleActions {
    padding: 0 0 1em 0;
    text-align: center;
    }
body[dir="ltr"] #revertButton:after {
    content: '\2009\f061';
    font-family: FontAwesome;
    font-style: normal;
    font-weight: normal;
    line-height: 1;
    vertical-align: baseline;
    display: inline-block;
    }
body[dir="rtl"] #revertButton:after {
    content: '\2009\f060';
    font-family: FontAwesome;
    font-style: normal;
    font-weight: normal;
    line-height: 1;
    vertical-align: baseline;
    display: inline-block;
    }
body[dir="ltr"] #commitButton:before {
    content: '\f060\2009';
    font-family: FontAwesome;
    font-style: normal;
    font-weight: normal;
    line-height: 1;
    vertical-align: baseline;
    display: inline-block;
    }
body[dir="rtl"] #commitButton:before {
    content: '\f061\2009';
    font-family: FontAwesome;
    font-style: normal;
    font-weight: normal;
    line-height: 1;
    vertical-align: baseline;
    display: inline-block;
    }
#revertButton,
#commitButton,
#diff.edit #editEnterButton {
    opacity: 0.25;
    pointer-events: none;
    }
#editStopButton,
#editCancelButton {
    display: none;
    }
#diff.dirty:not(.edit) #revertButton,
#diff.dirty:not(.edit) #commitButton {
    opacity: 1;
    pointer-events: auto;
    }
#diff.edit #editStopButton,
#diff.edit #editCancelButton {
    display: initial;
    }
#diff.edit #importButton,
#diff.edit #exportButton {
    display: none;
    }
#diff ul {
    border: 0;
    border-top: 1px solid #eee;
    list-style-type: none;
    margin: 0;
    overflow: hidden;
    padding: 1em 0 0 0;
    }
#diff.edit .right ul {
    visibility: hidden;
    }
#diff .left {
    border-right: 1px solid #eee;
    }
#diff .right > ul {
    color: #888;
    }
#diff li {
    background-color: #ddd;
    direction: ltr;
    padding: 0;
    text-align: left;
    white-space: nowrap;
    padding-left: 3px; /* a bit of padding; must also be in textarea */
    }
#diff li:nth-child(even) {
    background-color: #eee;
    }
#diff .right li {
    opacity: 0.5;
    }
#diff .right li:hover {
    }
#diff .right li.notLeft {
    color: #000;
    opacity: 1;
    }
#diff .right li.notRight {
    color: #000;
    }
#diff .right li.toRemove {
    color: #000;
    text-decoration: line-through;
    opacity: 1;
    }
#diff textarea {
    background-color: #f8f8ff;
    border: 0;
    border-top: 1px solid #eee;
    box-sizing: border-box;
    color: black;
    direction: ltr;
    font: inherit;
    height: 100%;
    left: 0;
    margin: 0;
    overflow: hidden;
    overflow-y: auto;
    padding: 1em 0 0 3px; /* same left and top padding as ul/li */
    position: absolute;
    resize: none;
    top: 0;
    visibility: hidden;
    white-space: pre; /* this implies nowrap; break only on \n and <br>.
                         nowrap doesn't consistently
                         respect \n's (example: Safari) per the CSS spec:
                         http://www.w3.org/wiki/CSS/Properties/white-space */
    width: 100%;
    word-wrap: normal;
    }
#diff.edit textarea {
    visibility: visible;
    }
